#include <PalmOS.h>
#include "Hexplode.h"

struct board {
	signed short value;
	signed char npos;	// Number of positive cells
	signed char nneg;	// Number of negative cells
	signed char moves;	// Number of moves played
	signed char n[5][5];
	signed char winner;
	signed short current_player;
};

UInt32 sound_level;

signed short showing = 1;

struct board game,previous;

int pre_level = 1;
int skill_level = 3;
int human = 2;
//int current_player = 1;

int last_i = -1,last_j = -1;
int undo_ok = 0;

signed char max[5][5] = {
	3,4,4,4,2,
	4,6,6,6,4,
	4,6,6,6,4,
	4,6,6,6,4,
	2,4,4,4,3
};

static int h_abs(int x) {
	return x<0 ? -x : x;
}

static int h_sign(int x) {
	return x<0 ? -1 : (x>0 ? 1 : 0);
}

/*
 * I have amortised the cost of evaluating positions by maintaining
 * nneg,npos and value.  Hence BoardEvaluate() is almost trivial
 */
static int BoardEvaluate(struct board *b,int s) {
	int score = 0;

	if (b->nneg==0)
		return s*10000;
	if (b->npos==0)
		return -s*10000;


	return s*b->value;
}

struct board level[6];
int next = 0;

static struct board *board_malloc() {
	return level+next++;
}

static void board_free() {
	--next;
}

static struct board *board_copy(struct board *b) {
	struct board *nb;

	*(nb = board_malloc()) = *b;

	return nb;
}

static void board_clear(struct board *b) {
	int i,j;
	b->npos = 0;
	b->nneg = 0;
	b->moves = 0;
	b->value = 0;
	b->winner = 0;
	for (i = 0; i<5; ++i)
		for (j = 0; j<5; ++j)
			b->n[i][j] = 0;
	b->current_player = 1;
}

#define A 10
#define B 5
#define C 10

RGBColorType background = { 0,0,255,0};
RGBColorType red = { 0,255,80,80};
RGBColorType blue = { 0,80,80,255};
RGBColorType black = { 0,0,0,0};
RGBColorType white = { 0,255,255,255};
Boolean color = 1;

static void EraseHex(int i,int j) {
	int centreX,centreY;
	int l;
	static int span[20] = {
		0,1,1,1,2,2,3,3,4,4,4,4,3,3,2,2,1,1,1,0
	};
		
	centreX = 70+(i-j)*(-A-B);
	centreY = 40+(i+j)*(C);

	for (l = 0; l<20; ++l)
		WinEraseLine(centreX-span[l],centreY+l,centreX+A+span[l],centreY+l);
		
}

static void DrawHex(int i,int j) {
	int centreX,centreY;
	
	centreX = 70+(i-j)*(-A-B);
	centreY = 40+(i+j)*(C);
	
	WinDrawLine(centreX,centreY,centreX+A,centreY);
	WinDrawLine(centreX+A,centreY,centreX+A+B,centreY+C);
	WinDrawLine(centreX+A+B,centreY+C,centreX+A,centreY+2*C);
	WinDrawLine(centreX+A,centreY+2*C,centreX,centreY+2*C);
	WinDrawLine(centreX,centreY+2*C,centreX-B,centreY+C);
	WinDrawLine(centreX-B,centreY+C,centreX,centreY);

}

//
// Draw drop shadow beneath entire board
//
static void DrawShadow() {
	int centreX,centreY,i;
	
	if (color) {
		WinSetForeColor(WinRGBToIndex(&black));
	}

	for (i = 0; i<4; ++i) {
		centreX = 70+(4-i)*(-A-B);
		centreY = 40+(4+i)*(C)+1;
		WinDrawLine(centreX+A,centreY+2*C,centreX,centreY+2*C);
		WinDrawLine(centreX,centreY+2*C,centreX-B,centreY+C);
	}
	for (i = 0; i<4; ++i) {
		centreX = 70+(i-4)*(-A-B);
		centreY = 40+(i+4)*(C)+1;
		WinDrawLine(centreX+A,centreY+2*C,centreX,centreY+2*C);
		WinDrawLine(centreX+A+B,centreY+C,centreX+A,centreY+2*C);
	}
	
	centreX = 70+(4-4)*(-A-B);
	centreY = 40+(4+4)*(C)+1;
	WinDrawLine(centreX+A,centreY+2*C,centreX,centreY+2*C);
	WinDrawLine(centreX,centreY+2*C,centreX-B,centreY+C);
	WinDrawLine(centreX+A+B,centreY+C,centreX+A,centreY+2*C);
}

static void FlashCell(int i,int j) {
	int k;
	int centreX,centreY;
	SndCommandType sndCmd;
	
	centreX = 70+(i-j)*(-A-B);
	centreY = 40+(i+j)*(C);
	
	sndCmd.cmd = sndCmdFreqDurationAmp;
	sndCmd.param1 = 500;
	sndCmd.param2 = 100;
	sndCmd.param3 = sound_level;
	SndDoCmd(NULL,&sndCmd,0);
	
	for (k =0; k<6;++k) {
		SysTaskDelay(20);
		WinInvertLine(centreX,centreY,centreX+A,centreY);
		WinInvertLine(centreX+A,centreY,centreX+A+B,centreY+C);
		WinInvertLine(centreX+A+B,centreY+C,centreX+A,centreY+2*C);
		WinInvertLine(centreX+A,centreY+2*C,centreX,centreY+2*C);
		WinInvertLine(centreX,centreY+2*C,centreX-B,centreY+C);
		WinInvertLine(centreX-B,centreY+C,centreX,centreY);
	}
}

static void DrawCell(int i,int j) {
	int centreX,centreY;
	char no[3];
	RectangleType erase;
	RGBColorType col = { 0,0,0,0 };
	
	centreX = 70+(i-j)*(-A-B);
	centreY = 40+(i+j)*(C);
	
	if (color) {
		if (game.n[i][j]>0) {
			WinSetBackColor(WinRGBToIndex(&red));
		} else if (game.n[i][j]<0) {
			WinSetBackColor(WinRGBToIndex(&blue));
		} else {
			RGBColorType col = { 0,255,255,255 };
			WinSetBackColor(WinRGBToIndex(&col));
		}
		EraseHex(i,j);
		WinSetForeColor(WinRGBToIndex(&col));
	}
	DrawHex(i,j);
	
	erase.topLeft.x = centreX;
	erase.topLeft.y = centreY+1;
	erase.extent.x = A;
	erase.extent.y = 2*C-2;
	WinEraseRectangle(&erase,0);
	
	if (game.n[i][j]) {
		RGBColorType col = { 0,0,0,0 };
		no[0] = game.n[i][j]>=0 ? ' ' : '-';
		no[1] = '0'+h_abs(game.n[i][j]);
		if (color)
			WinSetTextColor(WinRGBToIndex(&col));
		WinDrawChars(no,2,centreX+2,centreY+4);
	}
}

static void DrawBoard() {
	int i,j;

	for (i = 0; i<5; ++i)
		for (j = 0; j<5; ++j)
			DrawCell(i,j);
			
	DrawShadow();
}


/*
 * Explode cell at i,j
 *
 * This is where the CPU action lies.
 * XXX: Optimise more!!!
 *
 * Returns the winner if any:
 * -1 means negative has won
 * 1 means positive has won
 */
static int Hexplode(struct board *b,short s,short i,short j) {
	static char ij[256];
	register char *ptr;
	int n,d,t,u,v,m;

	ij[0] = i;
	ij[1] = j;
	
	ptr = ij+2;
	
	++b->moves;
	
	v = b->value;
	
	while (ptr>ij) {
		register int i,j;
		
		j = *--ptr;
		i = *--ptr;
		
		n = b->n[i][j];
		if (n<0)
			--b->nneg;
		if (n>0)
			--b->npos;
			
		d = s;
		if (s<0 && n>0 || s>0 && n<0)
			d = d-n-n;
		
		v += d;
		n += d;
		
		if (n<0)
			++b->nneg;
		if (n>0)
			++b->npos;
		
		if (showing)
			DrawCell(i,j);

		b->n[i][j] = n;
		
		if (b->moves>1) {
			b->value = v;
			if (b->npos==0)
				return b->winner = -1;
				
			if (b->nneg==0)
				return b->winner = 1;
		}

		m = max[i][j];		
		if (s>0 && n!=m || s<0 && n!=-m)
			continue;
	
		v -= n;
	
		if (n<0)
			--b->nneg;
		else if (n>0)
			--b->npos;
	
		b->n[i][j] = 0;

		t = i+1;
		u = j+1;
		
		if (t>=0 && t<=4) {
		
			if (j>=0 && j<=4) {
				*ptr++ = t;
				*ptr++ = j;
			}
			
			if (u>=0 && u<=4) {
				*ptr++ = t;
				*ptr++ = u;
			}
		}
		
		if (i>=0 && u>=0 && i<=4 && u<=4) {
			*ptr++ = i;
			*ptr++ = u;
		}

		t = i-1;
		if (t>=0 && j>=0 && t<=4 && j<=4) {
			*ptr++ = t;
			*ptr++ = j;
		}
		
		u = j-1;		
		
		if (u>=0 && u<=4) {			
			if (t>=0 && t<=4) {
				*ptr++ = t;
				*ptr++ = u;
			}
	
			if (i>=0 && i<=4) {
				*ptr++ = i;
				*ptr++ = u;
			}
		}
							
	}
	
	return 0;
}


/*
 * The strategy we use is a pretty pure minimax algorithm with alpha/beta
 * pruning.  The only heuristic used is to perform a low depth pre-evaluation
 * of each move in order to sort potential moves and then reanalyse those
 * positions to greater depth in order from best to worst.
 */

/*
 * So what's it worth to s to play (i,j)?
 * I'm assuming I can already get a score of alpha by another path
 * so if my opponent has any way to get my score
 * to less than that I need to check no further
 */
static int alphabeta(struct board *b,int s,int i,int j,int depth,int alpha,int beta) {
	int score;

	b = board_copy(b);

	Hexplode(b,s,i,j);

	if (b->nneg==0 && s<0 || b->npos==0 && s>0) {
		board_free();
		return -10000;
	}

	if (b->nneg==0 && s>0 || b->npos==0 && s<0) {
		board_free();
		return 10000;
	}

	if (depth<=0) {
		score = BoardEvaluate(b,s);
		board_free();
		return score;
	}

	/*
	 * Consider all of the opponents moves
	 */
	if (s>0) {
		for (i = 0; i<5; ++i)
			for (j = 0; j<5; ++j) {
				if (b->n[i][j]<=0) {
					score = -alphabeta(b,-s,i,j,depth-1,-beta,-alpha);
					if (score<beta)
						beta = score;
					if (beta<=alpha) {
						board_free();
						return beta;
					}
				}
			}
	} else {
		for (i = 0; i<5; ++i)
			for (j = 0; j<5; ++j) {
				if (b->n[i][j]>=0) {
					score = -alphabeta(b,-s,i,j,depth-1,-beta,-alpha);
					if (score<beta)
						beta = score;
					if (beta<=alpha) {
						board_free();
						return beta;
					}
				}
			}
	}

	board_free();

	return beta;
}

struct move {
	int i,j,score;
};

/*
 * XXX: Pointers are faster to work with but this function isn't
 * a major CPU consumer so leave for now.
 *
 * Partitioning function used by quicksort algorithm.
 */
static int partition(struct move *list,int n) {
	struct move pivot = list[n-1];
	int bottom = -1;
	int top = n-1;
	int done;

	done = 0;

	while (!done) {
		while (!done) {
			++bottom;
			if (bottom==top) {
				done = 1;
				break;
			}
			if (list[bottom].score<pivot.score) {
				list[top] = list[bottom];
				break;
			}
		}
		while (!done) {
			--top;
			if (top==bottom) {
				done = 1;
				break;
			}
			if (list[top].score>pivot.score) {
				list[bottom] = list[top];
				break;
			}
		}
	}

	list[top] = pivot;

	return top;
}

/*
 * Quicksort roughly evaluated positions
 */
static void move_sort(struct move *d,int n) {
	int r;
	if (n<=1)
		return;
	r = partition(d,n);
	move_sort(d,r);
	move_sort(d+r+1,n-r-1);
}

static void DisplayPercent(int p) {
	char msg[12];
	StrIToA(msg,p);
	StrCat(msg,"%...   ");
	if (color) {
		WinSetTextColor(WinRGBToIndex(&black));
		WinSetBackColor(WinRGBToIndex(&white));
	}
	WinDrawChars(msg,StrLen(msg),10,40);
}

static void ClearPercent() {
	WinDrawChars("              ",14,10,40);
}

//
// Main driver for alpha/beta search.
//
static int best(int *best_i,int *best_j,struct board *b,int s,int depth1,int depth2) {
	int best_score = -32767;
	int i,j,score;

	static struct move moves[25];
	int mptr = 0;
	
	int p = 0;

	for (i = 0; i<5; ++i)
		for (j = 0; j<5; ++j) {
			if (b->n[i][j]>=0 && s>0 || b->n[i][j]<=0 && s<0) {
				moves[mptr].i = i;
				moves[mptr].j = j;
				moves[mptr].score = alphabeta(b,s,i,j,depth1,-32767,32767);
				++mptr;
			}
			++p;
			DisplayPercent(p);
		}

	move_sort(moves,mptr);

	for (i = 0; i<mptr; ++i) {
		score = alphabeta(b,s,moves[i].i,moves[i].j,depth2,best_score,32767);
		if (score>best_score) {
			best_score = score;
			*best_i = moves[i].i;
			*best_j = moves[i].j;
			if (best_score>=10000)
				break;
		}
		p += 3;
		DisplayPercent(p);
	}
	DisplayPercent(100);
	
	return best_score;
}

typedef struct {
	UInt32 signature;
	struct board game;
	/*
	 * human:
	 * 00 both computer
	 * 10 negative human
	 * 01 positive human
	 * 11 both human
	 */
	int human;
	int pre_level;
	int skill_level;
//	int current_player;
} StarterPreferenceType;

typedef struct {
	UInt8 replaceme;
} StarterAppInfoType;

typedef StarterAppInfoType *StarterAppInfoPtr;

#define appFileCreator				 'STRT'
#define appVersionNum              0x01
#define appPrefID                  0x00
#define appPrefVersionNum          0x01

// Define the minimum OS version we support (2.0 for now).
#define ourMinVersion	sysMakeROMVersion(2,0,0,sysROMStageRelease,0)

static Err RomVersionCompatible(UInt32 requiredVersion,UInt16 launchFlags) {
	UInt32 romVersion;

	// See if we're on in minimum required version of the ROM or later.
	FtrGet(sysFtrCreator,sysFtrNumROMVersion,&romVersion);
	if (romVersion<requiredVersion)
		{
		if ((launchFlags & (sysAppLaunchFlagNewGlobals | sysAppLaunchFlagUIApp)) ==
			(sysAppLaunchFlagNewGlobals | sysAppLaunchFlagUIApp)) {
			FrmAlert (RomIncompatibleAlert);
		
			// Palm OS 1.0 will continuously relaunch this app unless we switch to 
			// another safe one.
			if (romVersion<ourMinVersion) {
				AppLaunchWithCommand(sysFileCDefaultApp,sysAppLaunchCmdNormalLaunch,NULL);
				}
			}
		
		return sysErrRomIncompatible;
		}

	return errNone;
}


/***********************************************************************
 *
 * FUNCTION:    GetObjectPtr
 *
 * DESCRIPTION: This routine returns a pointer to an object in the current
 *              form.
 *
 * PARAMETERS:  formId - id of the form to display
 *
 * RETURNED:    void *
 *
 * REVISION HISTORY:
 *
 *
 ***********************************************************************/
static void * GetObjectPtr(UInt16 objectID) {
	FormPtr frmP;

	frmP = FrmGetActiveForm();
	return FrmGetObjectPtr(frmP,FrmGetObjectIndex(frmP,objectID));
}


/***********************************************************************
 *
 * FUNCTION:    MainFormInit
 *
 * DESCRIPTION: This routine initializes the MainForm form.
 *
 * PARAMETERS:  frm - pointer to the MainForm form.
 *
 * RETURNED:    nothing
 *
 * REVISION HISTORY:
 *
 *
 ***********************************************************************/
static void MainFormInit(FormPtr frmP) {
}

static int InCell(int i,int j,int x,int y) {
	int cx = 70+(i-j)*(-A-B)+A/2;
	int cy = 40+(i+j)*(C)+C;
	
	#define R 10
	
	if ((x-cx)*(x-cx)+(y-cy)*(y-cy)<R*R)
		return 1;
	else
		return 0;
}

static int WhichCell(int *pi,int *pj,int x,int y) {
	int i,j;
	
	for (i = 0; i<5; ++i)
		for (j = 0; j<5; ++j) {
			if (InCell(i,j,x,y)) {
				*pi = i;
				*pj = j;
				
				return 1;
			}
		}
		
	return 0;
}

/***********************************************************************
 *
 * FUNCTION:    DoMove
 *
 * DESCRIPTION: Computes and displays best move for player
 *
 * PARAMETERS:  player - player whose move is computed
 *
 * RETURNED:    nothing
 *
 * REVISION HISTORY:
 *
 *
 ***********************************************************************/
static void DoMove(int player) {
	int i,j,r;
	
	showing = 0;
	best(&i,&j,&game,player,pre_level,skill_level);
	showing = 1;
	FlashCell(i,j);
	ClearPercent();
	last_i = i;
	last_j = j;
	r = Hexplode(&game,player,i,j);
	if (r==player)
		FrmAlert(IWinAlert);
	
	DrawBoard();
}

/***********************************************************************
 *
 * FUNCTION:    MainFormDoCommand
 *
 * DESCRIPTION: This routine performs the menu command specified.
 *
 * PARAMETERS:  command  - menu item id
 *
 * RETURNED:    nothing
 *
 * REVISION HISTORY:
 *
 *
 ***********************************************************************/
static Boolean MainFormDoCommand(UInt16 command) {
	Boolean handled = false;
	FormPtr frmP;
	UInt16 tmp;
	int last_human;
	int i,j;
	int s;
//	int current_player = 1;

	switch (command) {
		case HelpAbout:
			MenuEraseStatus(0);
			frmP = FrmInitForm (AboutForm);
			FrmDoDialog (frmP);					// Display the About Box.
			FrmDeleteForm (frmP);
			handled = true;
			break;

		case HelpRules:
			MenuEraseStatus(0);
			FrmHelp(RulesString);
			handled = true;
			break;

		case HelpHelp:
			MenuEraseStatus(0);
			FrmHelp(HelpString);
			handled = true;
			break;

		case MainOptionsRegister:
			MenuEraseStatus(0);
			FrmHelp(RegistrationString);
			handled = true;
			break;

		/*
		 * positive moves first.
		 */
		case GameNewGame:
			MenuEraseStatus(0);
			board_clear(&game);
			undo_ok = 0;
			last_i = -1;
			last_j = -1;
//			game.current_player = 1;
			DrawBoard();
			/* 
			 * If the first player's not a human then
			 * the computer should move first.
			 */
			if (!(human & 2)) {
				DoMove(1);
				game.current_player = -1;
			}
//			if (human==-1) {
				handled = true;
//			}
//			printf("New game,player = %d\n",game.current_player);
			break;

		case GameUndo:
			MenuEraseStatus(0);
			if (undo_ok) {
				game = previous;
				DrawBoard();
				last_i = -1;
				last_j = -1;
			} else
				FrmAlert(YouCantUndoAlert);
			handled = true;
			break;
			
		case GameHint:
			MenuEraseStatus(0);
			showing = 0;
			s = best(&i,&j,&game,game.current_player,pre_level,skill_level);
			if (s==10000) {
				FrmAlert(YouCanWinAlert);
				DrawBoard();
			} else if (s==-10000) {
				FrmAlert(YouCantWinAlert);
				DrawBoard();
			}
			showing = 1;
			FlashCell(i,j);
			last_i = i;
			last_j = j;
			ClearPercent();
			break;
			
		case GameShowLastMove:
			MenuEraseStatus(0);
			if (last_i>=0)
				FlashCell(last_i,last_j);
			else
				FrmAlert(NoLastMoveAlert);
			handled = true;
			break;

		/*
		 * Handle difficulty level and which player is which.
		 */
		case MainOptionsPreferences:
			MenuEraseStatus(0);
			frmP = FrmInitForm(PreferencesForm);
			switch (skill_level) {
			case 1:
				tmp = PreferencesEasyPushButton;
				break;
			case 3:
				tmp = PreferencesMediumPushButton;
				break;
			case 5:
				tmp = PreferencesDifficultPushButton;
				break;
			}
			FrmSetControlGroupSelection ( frmP,PreferencesFormGroupID,tmp );
			switch (human) {
			case 3:
				tmp = PreferencesBothPushButton;
				break;
			case 2:
				tmp = PreferencesPositivePushButton;
				break;			
			case 1:
				tmp = PreferencesNegativePushButton;
				break;			
			}
			FrmSetControlGroupSelection ( frmP,PreferencesFormGroupID2,tmp );
			last_human = human;
			FrmDoDialog(frmP);
			tmp = FrmGetObjectId(frmP,FrmGetControlGroupSelection(frmP,PreferencesFormGroupID));
				switch (tmp) {
				case PreferencesEasyPushButton:
					skill_level = 1;
					pre_level = 1;
					break;
				case PreferencesMediumPushButton:
					skill_level = 3;
					pre_level = 1;
					break;
				case PreferencesDifficultPushButton:
					skill_level = 5;
					pre_level = 3;
					break;
			}
			tmp = FrmGetObjectId(frmP,FrmGetControlGroupSelection(frmP,PreferencesFormGroupID2));
				switch (tmp) {
				case PreferencesBothPushButton:
					human = 3;
					break;
				case PreferencesPositivePushButton:
					human = 2;
					break;
				case PreferencesNegativePushButton:
					human = 1;
					break;
			}
			FrmDeleteForm(frmP);
//			printf("human=%d\n",human);
#if 0
			if (human!=last_human) {
				DoMove(last_human);
				undo_ok = 0;
			}
#endif
			if (game.current_player==-1 && !(human & 1) ||
				game.current_player==1 && !(human & 2)) {
				DrawBoard();
				DoMove(game.current_player);
				undo_ok = 0;
				game.current_player = -game.current_player;
			}

			handled = true;
			break;

		}
	DrawBoard();
	
	return handled;
}


/***********************************************************************
 *
 * FUNCTION:    MainFormHandleEvent
 *
 * DESCRIPTION: This routine is the event handler for the 
 *              "MainForm" of this application.
 *
 * PARAMETERS:  eventP  - a pointer to an EventType structure
 *
 * RETURNED:    true if the event has handle and should not be passed
 *              to a higher level handler.
 *
 * REVISION HISTORY:
 *
 *
 ***********************************************************************/
static Boolean MainFormHandleEvent(EventPtr eventP) {
   Boolean handled = false;
   FormPtr frmP;

	switch (eventP->eType) 
		{
		case menuEvent:
			return MainFormDoCommand(eventP->data.menu.itemID);

		case frmOpenEvent:
			frmP = FrmGetActiveForm();
			MainFormInit( frmP);
			FrmDrawForm ( frmP);
			
			DrawBoard();
			
			handled = true;
			break;
			
		case penDownEvent:
		{	
			int i,j,r;
//			printf("Pen down for %d\n",game.current_player);
			if (WhichCell(&i,&j,eventP->screenX,eventP->screenY) && game.n[i][j]*game.current_player>=0 && !game.winner) {
//				printf("current = %d\n",current_player);
//				printf("Legal\n");
				previous = game;
				undo_ok = 1;
//				printf("flash\n");
				FlashCell(i,j);
//				printf("Hexplode for %d\n",game.current_player);
				r = Hexplode(&game,game.current_player,i,j);
//				printf("Hexploded\n");
				if (r==game.current_player) {
					FrmAlert(YouWinAlert);
					DrawBoard();
				} else {
					game.current_player = -game.current_player;
					if (game.current_player==-1 && !(human & 1) ||
						game.current_player==1 && !(human & 2)) {
//						printf("Computer...\n");
						DrawBoard();
						DoMove(game.current_player);
						game.current_player = -game.current_player;
//						printf("...done\n");
					}
					DrawBoard();
				}
			}
//			printf("Move over\n");
		}
			break;

		default:
			break;
		
		}
	
	return handled;
}

/***********************************************************************
 *
 * FUNCTION:    AppHandleEvent
 *
 * DESCRIPTION: This routine loads form resources and set the event
 *              handler for the form loaded.
 *
 * PARAMETERS:  event  - a pointer to an EventType structure
 *
 * RETURNED:    true if the event has handle and should not be passed
 *              to a higher level handler.
 *
 * REVISION HISTORY:
 *
 *
 ***********************************************************************/
static Boolean AppHandleEvent(EventPtr eventP) {
	UInt16 formId;
	FormPtr frmP;

	if (eventP->eType == frmLoadEvent)
		{
		// Load the form resource.
		formId = eventP->data.frmLoad.formID;
		frmP = FrmInitForm(formId);
		FrmSetActiveForm(frmP);

		// Set the event handler for the form.  The handler of the currently
		// active form is called by FrmHandleEvent each time is receives an
		// event.
		switch (formId)
			{
			case MainForm:
				FrmSetEventHandler(frmP,MainFormHandleEvent);
				break;

			default:
//				ErrFatalDisplay("Invalid Form Load Event");
				break;

			}
		return true;
		}
	
	return false;
}


/***********************************************************************
 *
 * FUNCTION:    AppEventLoop
 *
 * DESCRIPTION: This routine is the event loop for the application.  
 *
 * PARAMETERS:  nothing
 *
 * RETURNED:    nothing
 *
 * REVISION HISTORY:
 *
 *
 ***********************************************************************/
static void AppEventLoop(void) {
	UInt16 error;
	EventType event;

	do {
		EvtGetEvent(&event,evtWaitForever);

		if (! SysHandleEvent(&event))
			if (! MenuHandleEvent(0,&event,&error))
				if (! AppHandleEvent(&event))
					FrmDispatchEvent(&event);

	} while (event.eType != appStopEvent);
}


/***********************************************************************
 *
 * FUNCTION:     AppStart
 *
 * DESCRIPTION:  Get the current application's preferences.
 *
 * PARAMETERS:   nothing
 *
 * RETURNED:     Err value 0 if nothing went wrong
 *
 * REVISION HISTORY:
 *
 *
 ***********************************************************************/
static Err AppStart(void) {
    StarterPreferenceType prefs;
    UInt16 prefsSize;

	WinScreenMode(winScreenModeGet,0,0,0,&color);

	// Read the saved preferences / saved-state information.
	prefsSize = sizeof(StarterPreferenceType);
	if (PrefGetAppPreferences(appFileCreator,appPrefID,&prefs,&prefsSize,true) != 
		noPreferenceFound && prefs.signature==670) {
			game = prefs.game;
			human = prefs.human;
			pre_level = prefs.pre_level;
			skill_level = prefs.skill_level;
//				current_player = prefs.current_player;
//				printf("1.cp=%d\n",game.current_player);
		} else {
			board_clear(&game);
//				printf("2.cp=%d\n",game.current_player);
		}
			
		last_i =-1;
		last_j = -1;
		undo_ok = 0;

	//
	// Get sound preferences
	//
	sound_level = PrefGetPreference(prefGameSoundVolume);
	
   return errNone;
}


/***********************************************************************
 *
 * FUNCTION:    AppStop
 *
 * DESCRIPTION: Save the current state of the application.
 *
 * PARAMETERS:  nothing
 *
 * RETURNED:    nothing
 *
 * REVISION HISTORY:
 *
 *
 ***********************************************************************/
static void AppStop(void) {
   StarterPreferenceType prefs;

	// Write the saved preferences / saved-state information.  This data 
	// will be backed up during a HotSync.
	prefs.signature = 670;
	prefs.game = game;
	prefs.human = human;
	prefs.skill_level = skill_level;
	prefs.pre_level = pre_level;
//	prefs.current_player = current_player;
	PrefSetAppPreferences (appFileCreator,appPrefID,appPrefVersionNum,
		&prefs,sizeof (prefs),true);
		
	// Close all the open forms.
	FrmCloseAllForms ();
}


/***********************************************************************
 *
 * FUNCTION:    StarterPalmMain
 *
 * DESCRIPTION: This is the main entry point for the application.
 *
 * PARAMETERS:  cmd - word value specifying the launch code. 
 *              cmdPB - pointer to a structure that is associated with the launch code. 
 *              launchFlags -  word value providing extra information about the launch.
 *
 * RETURNED:    Result of launch
 *
 * REVISION HISTORY:
 *
 *
 ***********************************************************************/
static UInt32 StarterPalmMain(UInt16 cmd,MemPtr cmdPBP,UInt16 launchFlags) {
	Err error;

	error = RomVersionCompatible (ourMinVersion,launchFlags);
	if (error) return (error);

	switch (cmd) {
		case sysAppLaunchCmdNormalLaunch:
			error = AppStart();
			
			if (error) {
				return error;
			}
				
			FrmGotoForm(MainForm);
			AppEventLoop();
			AppStop();
			break;

		default:
			break;
		}
	
	return errNone;
}


/***********************************************************************
 *
 * FUNCTION:    PilotMain
 *
 * DESCRIPTION: This is the main entry point for the application.
 *
 * PARAMETERS:  cmd - word value specifying the launch code. 
 *              cmdPB - pointer to a structure that is associated with the launch code. 
 *              launchFlags -  word value providing extra information about the launch.
 * RETURNED:    Result of launch
 *
 * REVISION HISTORY:
 *
 *
 ***********************************************************************/
UInt32 PilotMain(UInt16 cmd,MemPtr cmdPBP,UInt16 launchFlags) {
    return StarterPalmMain(cmd,cmdPBP,launchFlags);
}

